unit krScramDisk;

(******************************************************************************
        TkrScramDisk - a Delphi 32-bit component for controlling ScramDisk
            ScramDisk - Virtual disk encryption for Windows 95/98

 ScramDisk was written by Aman and is available from
 http://www.hertreg.ac.uk/ss/

 The TkrScramDisk Delphi Component is 1998 Kwik-Rite Development and was
 written by Andy Jeffries - http://www.kwikrite.clara.net/

 Changes/corrections to this component are welcome, but they should be sent to
 Kwik-Rite Development so they can be incorporated in to the next official
 release.

 Version History
 ------- -------

 Version    Changes
 0.9 Beta   First version released to the public by Kwik-Rite Development
 ******************************************************************************)

interface

uses
  Classes, SysUtils, Windows, krSHA, krSdStructures;

type
  TkrScramDisk = class(TComponent)
  private
    hScramDiskVxD : THandle;
    FActive : Boolean;
    pThePasswords : Pointer;
    {FWavInfo : TWavInfo;}
  protected
    function EjectStop(DriveLetter : string; EjectFunction : Boolean) : Boolean;virtual;
    procedure Connect;virtual;
    procedure Disconnect;virtual;
    procedure SetActive(AValue : Boolean);virtual;
    procedure UpdateDeviceDriverTime;virtual;
    procedure GetPasswordBufferInt;
    procedure PassPasswords;virtual;
    function CheckWav(Filename : string) : Integer;virtual;
    function FindWavData(fs : TFileStream) : Integer;virtual;
    function IOCtlLock(DriveNo : Integer; Permissions : Integer; FunctionNo : Integer) : Integer;virtual;
    function LockLogDrive(DriveNo : Integer; Mode : Integer) : Integer;virtual;
    function LockDrive(Drive : string; Mode : Integer) : Boolean;virtual;
    procedure BrutalDismountSlot(SlotNo: Integer);virtual;
  public
    constructor create(AOwner : TComponent); override;
    destructor destroy; override;
    function GetSlotInfo(SlotNo : Integer) : TSlotInfo;virtual;
    procedure UpdateSlotInfo;virtual;
    procedure Dismount(SlotNo : Integer; Brutal : Boolean = true); overload;virtual;
    procedure Dismount(Drive : string; Brutal : Boolean = true); overload;virtual;
    procedure RefreshApplication;virtual;
    function MountContainer(Filename : string) : Boolean;virtual;
    procedure MountPartitions;virtual;
    function SetPasswordsRedScreen(ScreenTitle : string) : Boolean;virtual;
    procedure SetPasswords(Password1, Password2, Password3, Password4 : string);virtual;
    procedure ClearPasswords;virtual;
    procedure MountAll;virtual;
  published
    property Active : Boolean read FActive write SetActive;
  end;

var
  MyCVS : array[0..7] of TCryptVol;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('KwikRite', [TkrScramDisk]);
end;

(******************************************************************************
                         TkrSCRAMDISK DEFINITIONS
 ******************************************************************************)

{ --- Standard component and connection routines --- }

constructor TkrScramDisk.Create(AOwner : TComponent);
begin
  inherited create(AOwner);
  FActive := False;
end;

destructor TkrScramDisk.Destroy;
begin
  if FActive then
    CloseHandle(hScramDiskVxD);
  inherited Destroy;
end;

procedure TkrScramDisk.Connect;
var
  pchDriver : PChar;
  iDriverVersion : ^Integer;
  dwBytesReturned : DWORD;
begin
  if not FActive then
    begin
      hScramDiskVxD := CreateFile('\\.\sd', 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
      if hScramDiskVxD = INVALID_HANDLE_VALUE then
        raise ESDVxdNotFound.Create('Scramdisk device driver not present'#13#10 +
          'You need to make sure the file ''SD.VXD'' is present'#13#10 +
          'in your ''<WINDIR>\SYSTEM\IOSUBSYS'' directory'#13#10 +
          'then restart Windows'#13#10#13#10 +
          'If you have just installed Scramdisk without'#13#10 +
          'restarting the computer you need to restart it now'#13#10 +
          'if you wish to use Scramdisk');
      DeviceIoControl(hScramDiskVxD, GETDRIVERVERSION, @pchDriver, 0,
        @iDriverVersion, 0, dwBytesReturned, nil);
      if iDriverVersion^ < $202 then
        raise ESDWrongVersion.Create('The installed version of the SD.VXD driver (in ' +
          'IOSUBSYS) is out of date'#13#10'Please reinstall Scramdisk with the ' +
          'correct version');
      FActive := True;
      GetPasswordBufferInt;
    end;
end;

procedure TkrScramDisk.Disconnect;
begin
  if FActive then
    CloseHandle(hScramDiskVxD);
  FActive := false;
end;

{ --- Information retrieval routines --- }

function TkrScramDisk.GetSlotInfo(SlotNo : Integer) : TSlotInfo;
var
  infoslot : array[0..7] of char;
  dwBytesReturned : DWORD;
  InfoStruct : TInfoStruct;
  Dummy : Char;
  SysTime : TSystemTime;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  infoslot[0] := char(SlotNo);
  DeviceIoControl(hScramDiskVxD, INQUIRE_SLOT, @infoslot, 0, nil, 0, dwBytesReturned, nil);

  Result.SlotNo := SlotNo;
  if (Ord(InfoSlot[1]) <= NUMBER_MOUNT_TYPES) and
    (Ord(InfoSlot[1]) > 1) then
    begin
      Result.MountType := TMountType(InfoSlot[1]);
      Result.HostDriveLetter := InfoSlot[2];
      Result.DriveMountedAs := Chr(MyCVS[SlotNo].drive + 65);
      Result.VolumeName := string(MyCVS[SlotNo].volname);
      Result.CipherType := CipherType[MyCVS[SlotNo].cipher];
      Result.FileName := MyCVS[SlotNo].mounted_file_name;
      InfoStruct.SlotNo := SlotNo;
      DeviceIoControl(hScramDiskVxD, READINFOBLOCK, @InfoStruct, 0, @Dummy, 0, dwBytesReturned, nil);
      Result.Comments := InfoStruct.infotext;
      Result.PreferredDrive := Char(65 + InfoStruct.preferredletter);
      Result.NoAccesses := InfoStruct.accessed;
      SysTime := InfoStruct.oldtime;
      try
      Result.LastAccess := EncodeDate(SysTime.wYear, SysTime.wMonth, SysTime.wDay) +
        EncodeTime(SysTime.wHour, SysTime.wMinute, SysTime.wSecond, SysTime.wMilliseconds);
      except
        ;
      end;
    end
  else
    Result.MountType := mtNotMounted;
end;

procedure TkrScramDisk.UpdateSlotInfo;
var
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  DeviceIoControl(hScramDiskVxD, PUTCVS, @MyCVS, 55, @MyCVS, 55, dwBytesReturned, nil);
end;

{ --- Mount routines --- }

function TkrScramDisk.MountContainer(Filename : string) : Boolean;
var
  fs : TMountFileStruct;
  mf : ^TMountFileStruct;
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  mf := @fs;
  mf.dcb := 0;
  mf.WavOffset := 0;
  mf.FileName := PChar(Filename);
  mf.Critical := nil; // SKF mount only

  {if CheckWav(Filename)=1 then  //  is it a 16 bit .WAV ?
          mf.WavOffset:=FWavInfo.FileOffset;}

  UpdateDeviceDriverTime(); // MAKE SURE DRIVER KNOWS THE TIME...
  DeviceIoControl(hScramDiskVxd, MOUNTFILE, mf, 0, mf, 0, dwBytesReturned, nil);

  {        if mf.mountresult<>3 then // not mounted...
          begin
          if mf.wavoffset<>0 then
                  begin
                  mf.wavoffset:=mf.wavoffset and $80000000;

                  UpdateDeviceDriverTime();				// MAKE SURE DRIVER KNOWS THE TIME...
                  DeviceIoControl(hScramDiskVxd, MOUNTFILE, mf, 0, mf, 0, dwBytesReturned, nil);
                  end;
          end;}


  Result := (mf.MountResult = 3);
  if Result then
    begin
      EjectStop(mf.FileName[0], true); // StopEject if file on removable media
      ClearPasswords;
      RefreshApplication; // Refresh the scramdisk application if it is running.
    end;
end;

procedure TkrScramDisk.MountAll;
var
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxDNotConnected.Create(E_NOT_CONNECTED);

  DeviceIoControl(hScramDiskVxD, MOUNTNEWDISK, nil, 0, nil, 0, dwBytesReturned, nil);
  RefreshApplication;
end;

procedure TkrScramDisk.MountPartitions;
var
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  DeviceIoControl(hScramDiskVxd, MOUNTNEWDISK, nil, 0, nil, 0, dwBytesReturned, nil);
end;

{ --- Dismount routines --- }

procedure TkrScramDisk.Dismount(Drive : string; Brutal : Boolean);
var
  SlotNo : Integer;
  ExamineSlot : Integer;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  //SlotInfo := GetSlotInfo(SlotNo);
  if not LockDrive(Drive, 1) then
    raise ESDDriveInUse.Create(E_DRIVE_IN_USE);

  SlotNo := -1;
  UpdateSlotInfo;
  for ExamineSlot := 0 to 7 do
    if GetSlotInfo(ExamineSlot).DriveMountedAs = Copy(Drive, 1, 1) then
      SlotNo := ExamineSlot;

  BrutalDismountSlot(SlotNo);
end;

procedure TkrScramDisk.Dismount(SlotNo : Integer; Brutal : Boolean);
var
  SlotInfo : TSlotInfo;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  try
    UpdateSlotInfo;
    SlotInfo := GetSlotInfo(SlotNo);
    if Brutal then
      if not LockDrive(SlotInfo.DriveMountedAs + ':', 1) then
        raise ESDDriveInUse.Create(E_DRIVE_IN_USE);

    BrutalDismountSlot(SlotNo);
  except
    on e : exception do ;
    { Likely to be an access violation, although the code appears to have worked}
  end;
end;

procedure TkrScramDisk.BrutalDismountSlot(SlotNo : Integer);
var
  dwBytesReturned : DWORD;
  CloseMode : Integer;
  LocalSlotNo : Integer; //local copy of SlotNo for passing to SD.VXD
  SlotInfo : TSlotInfo;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  CloseMode := 0; // preclose
  LocalSlotNo := SlotNo;
  DeviceIoControl(hScramDiskVxd, CLOSE_CRYPTED_DEVICE_N, @CloseMode, 0,
    @LocalSlotNo, 0, dwBytesReturned, nil);

  if (CloseMode <> 1) then
    begin
      CloseMode := 1;
      LocalSlotNo := SlotNo;
      DeviceIoControl(hScramDiskVxd, CLOSE_CRYPTED_DEVICE_N, @CloseMode, 0,
        @LocalSlotNo, 0, dwBytesReturned, nil);
    end;

  if (SlotInfo.MountType = mtSVLFile) or (SlotInfo.MountType = mtWAVFile) then
    EjectStop(SlotInfo.HostDriveLetter, False); // release eject if removable and file...
  RefreshApplication;
end;

{ --- WAV mounting functions --- }
{ The functions will be ready for the next release }

function TkrScramDisk.FindWavData(fs : TFileStream) : Integer;
{var
  i : ^Word;
  w : ^Byte;
  c : PChar;
  c2 : PChar;}
begin
  {  FWavInfo.WavType := $FFFF;
    FWavInfo.BitsPerSample := $FFFF;
    FWavInfo.NumChannels := $FFFF;
    Result := 0}
      (*if (cmps (buffer,"RIFF",4)==0) c=(unsigned char * ) buffer+8;
  else return -1;

  if (cmps((char * ) c,"WAVEfmt ",8)==0 ) c2=c+8;
  else return -1;

  w=(unsigned short * ) c2;
  wav->type=w[2];
  wav->numchannels=w[3];
  wav->bitspersample=w[9];

  i=(unsigned int * ) c2;
  c2+=4;
  c2+=*i;

  do
  {
  if ((char * )c2>buffer+64000) return -1;
  if (cmps((char * )c2,"data",4)==0) break;
  i=(unsigned int * )  c2+4;
  c2+=*i;
  c2+=8;
  }while(1);

  c2+=4;
  i=(unsigned int * )c2;
  wav->length=*i;
  c2+=4;
  wav->datapntr=(char * ) c2;
  wav->fileoffset=(char * )c2-buffer;
  return 0;

  }
  *)
  Result := 0;
end;

function TkrScramDisk.CheckWav(Filename : string) : Integer;
{var
  valid : Integer;
  TestFile : TFileStream;}
begin
  {  Result := 0;
    if Copy(FileName, Length(FileName) - 3, 3) = '.wav' then
      begin
        TestFile := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);
        try
          valid := FindWavData(TestFile);
        finally
          TestFile.Free;
        end;
        {if (FWavInfo.BitsPerSample = 16) and (FWavInfo.WavType = 1) and (FWavInfo.Length > (2048 * 1024)) then
          result := 1
        else}
  Result := -1;
  {end;}
end;

{ --- Password routines - to be called by the client application --- }

function TkrScramDisk.SetPasswordsRedScreen(ScreenTitle : string) : Boolean;
var
  dwBytesReturned : DWORD;
  messagestr : PChar;
  accepted : LongInt;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  messagestr := PChar(ScreenTitle);
  accepted := 256; //Pre-condition to ensure a result
  GetPasswordBufferInt;
  DeviceIoControl(hScramDiskVxD, GETINKEY, messagestr, 0,
    @accepted, 0, dwBytesReturned, nil);
  Result := (accepted = 0);
  PassPasswords;
end;

procedure TkrScramDisk.SetPasswords(Password1, Password2, Password3, Password4 : string);
var
  Passwords : TMemoryStream;
  Blank : Char;
  n : Integer;
begin
  Passwords := TMemoryStream.Create;
  try
    Blank := #0;
    for n := 0 to 159 do
      Passwords.Write(Blank, 1);
    Passwords.Position := 0;
    Passwords.Write(Password1[1], length(Password1));
    Passwords.Position := 40;
    Passwords.Write(Password2[1], length(Password2));
    Passwords.Position := 80;
    Passwords.Write(Password3[1], length(Password3));
    Passwords.Position := 120;
    Passwords.Write(Password4[1], length(Password4));
    GetPasswordBufferInt;
    CopyMemory(pThePasswords, Passwords.Memory, 160);
    PassPasswords;
  finally
    Passwords.Free;
  end;
end;

procedure TkrScramDisk.ClearPasswords;
const
  ObliteratePWord1 = 'THIS IS NOT';
  ObliteratePWord2 = 'A VALID';
  ObliteratePWord3 = 'SET OF';
  ObliteratePWord4 = 'PASSWORDS';
begin
  SetPasswords(ObliteratePWord1, ObliteratePWord2, ObliteratePWord3, ObliteratePWord4);
  SetPasswords('', '', '', '');
end;

{ --- Internal password routines - not to be called by the client application --- }

procedure TkrScramDisk.GetPasswordBufferInt;
var
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  DeviceIoControl(hScramDiskVxD, GETPASSWORDBUFFER, @pThePasswords, 0,
    @pThePasswords, 0, dwBytesReturned, nil);
end;

procedure TkrScramDisk.PassPasswords;
var
  dwBytesReturned : DWORD;
  dval : TSHA1Digest;
begin
  //Called internally...no need to check FActive
  dval := SHA1(pThePasswords, 160); // password HASH algortihm SHA 1
  UpdateDeviceDriverTime();

  // send both the password and SHA1 digest to the device driver
  DeviceIoControl(hScramDiskVxd, SUPPLY_PASSWORD, pThePasswords, 160, @dval, 20,
    dwBytesReturned, nil);

  RefreshApplication;
end;

{ --- General routines --- }

procedure TkrScramDisk.RefreshApplication;
var
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  DeviceIoControl(hScramDiskVxd, APP_REQUEST_REFRESH, nil, 0, nil, 0, dwBytesReturned, nil);
end;


{ --- Internal routines - not to be called by client applications --- }

function TkrScramDisk.EjectStop(DriveLetter : string; EjectFunction : Boolean) : Boolean;
var
  hVWin32VxD : THandle;
  Registers : TDeviceIOControlRegisters;
  EjectResult : Boolean;
  dwBytesReturned : DWORD;
  LockFunction : Integer;
  DriveNumber : Integer;
  ParamBlock : TParamBlock;
  DriveChar : Char;
begin
  DriveLetter := Uppercase(DriveLetter);
  DriveChar := DriveLetter[1];
  DriveNumber := (Ord(DriveChar) - 64);
  LockFunction := $48;

  if Ejectfunction = true then
    ParamBlock.Operation := 0 //lock
  else
    ParamBlock.Operation := 1; //unlock

  hVWin32VxD := CreateFile('\\.\vwin32', 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
  Registers.EAX := $440D;
  Registers.EBX := DriveNumber;
  Registers.ECX := $0800 or LockFunction;
  Registers.EDX := DWORD(@ParamBlock);
  Registers.Flags := $0001;
  EjectResult := DeviceIoControl(hVWin32Vxd, VWIN32_DIOC_DOS_IOCTL, @Registers, SizeOf(Registers),
    @Registers, SizeOf(Registers), dwBytesReturned, nil);
  if not EjectResult or ((Registers.Flags and $0001) <> 0) then
    CloseHandle(hVWin32Vxd);
  Result := (Registers.Flags and $1) = 0;
end;

procedure TkrScramDisk.UpdateDeviceDriverTime;
var
  s : TSystemTime;
  dwBytesReturned : DWORD;
begin
  if not FActive then
    raise ESDVxdNotConnected.Create(E_NOT_CONNECTED);

  GetLocalTime(s);
  DeviceIoControl(hScramDiskVxd, SUPPLYDATEANDTIME, @s, 0, @s, 0, dwBytesReturned, nil);
end;

procedure TkrScramDisk.SetActive(AValue : Boolean);
begin
  if AValue <> FActive then
    begin
      if AValue then
        Connect
      else
        Disconnect;
    end;
end;

function TkrScramDisk.IOCtlLock(DriveNo : Integer; Permissions : Integer; FunctionNo : Integer) : Integer;
var
  hDevice : THandle;
  Reg : TDeviceIOControlRegisters;
  cb : DWORD;
  lockfunc : Integer;
begin

  if FunctionNo <> 0 then
    lockfunc := $4A
  else
    lockfunc := $6A;

  hDevice := CreateFile('\\.\vwin32', 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);

  reg.EAX := $440D;
  reg.EBX := DriveNo;
  reg.ECX := $0800 or lockfunc;
  reg.EDX := permissions;
  reg.Flags := $0001;

  DeviceIoControl(hDevice, VWIN32_DIOC_DOS_IOCTL, @reg, sizeof(reg),
    @reg, sizeof(reg), cb, nil);

  CloseHandle(hDevice);

  Result := reg.Flags and 1;
end;


function TkrScramDisk.LockLogDrive(DriveNo, Mode : Integer) : Integer;
var
  a : Integer;
begin
  if (mode > 0) then
    begin
      IOCtlLock(DriveNo, 0, 1);
      a := IOCtlLock(DriveNo, 4, 1);
    end
  else
    begin
      a := 0;
      IOCtlLock(DriveNo, 0, 0);
      IOCtlLock(DriveNo, 0, 0);
    end;

  Result := a;
end;

function TkrScramDisk.LockDrive(Drive : string; Mode : Integer) : Boolean;
var
  DriveLett : Integer;
  a : Integer;
  UpCaseDrive : string;
begin
  UpCaseDrive := Uppercase(Drive);
  DriveLett := Ord(UpCaseDrive[1]);
  a := 1;
  DriveLett := DriveLett - 65; // 'A' in ASCII
  Inc(Drivelett);

  if ((drivelett > 1) and (drivelett < 27)) then
    a := LockLogDrive(drivelett, mode);
  Result := (a <> 1);
end;


end.

